from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.Util.Padding import pad, unpad
from Crypto.Signature import pss
from Crypto.PublicKey import RSA
from Crypto.Hash import SHA256, HMAC
from Crypto.Random import get_random_bytes
from enum import Enum
import json
import mindlakesdk.settings as settings
import logging

class ResultType:
    def __init__(self, code: int, message: str = None, data = None):
        self.code = code
        self.message = message
        self.data = data
        
    def __bool__(self):
        return self.code == 0
    
class DataType(Enum):
    int4 = 1
    int8 = 2
    float4 = 3
    float8 = 4
    decimal = 5
    text = 6
    timestamp = 7

class Session:
    def __init__(self) -> None:
        self.walletAddress = None
        self.chainID = None
        self.isRegistered = False
        self.mk = None
        self.sk = None
        self.isLogin = False
        self.token = None
        self.accountID = None
        self.nodePK = None
        self.pkID = None
        self.appKey = None
        self.gateway = None
        self.requstSession = None
        
class BlockChain:
    def __init__(self, chainObj: dict) -> None:
        self.chainID = chainObj["chainId"]
        self.rpcNode = chainObj["rpcNodeUrl"]
        self.contract = chainObj["smartAddress"]
        self.abi = chainObj["abi"]

    def getNameSpace(self, walletAddress: str):
        if self.value > 0:
            return f'evm{self.value}_{walletAddress}'

def genRSAKey():
    rsaKey = RSA.generate(2048)
    return rsaKey

def genAESKey():
    aesKey = get_random_bytes(16)
    return aesKey

def sha256Hash(data):
    h = SHA256.new(data)
    return h.digest()

def aesEncrypt(key, iv, data):
    cipher = AES.new(key, AES.MODE_CBC, iv)
    padded_data = pad(data, AES.block_size)
    encrypted_data = cipher.encrypt(padded_data)
    return encrypted_data

def aesDecrypt(key, iv, data):
    cipher = AES.new(key, AES.MODE_CBC, iv)
    decrypted_data = cipher.decrypt(data)
    result = unpad(decrypted_data, AES.block_size)
    return result

def aesGCMEncrypt(key, iv, data):
    cipher = AES.new(key, AES.MODE_GCM, iv)
    encrypted_data, tag = cipher.encrypt_and_digest(data)
    return tag + encrypted_data

def aesGCMDecrypt(key, iv, data, tag):
    cipher = AES.new(key, AES.MODE_GCM, iv)
    decrypted_data = cipher.decrypt_and_verify(data, tag)
    return decrypted_data

def hmacHash(key, data):
    h = HMAC.new(key, digestmod=SHA256)
    h.update(data)
    return h.digest()

def rsaEncrypt(pubKey, data):
    public_key = RSA.import_key(pubKey)
    cipher = PKCS1_OAEP.new(public_key, hashAlgo=SHA256)
    encrypted_data = cipher.encrypt(data)
    return encrypted_data

def rsaDecrypt(priKey, data):
    private_key = RSA.import_key(priKey)
    cipher = PKCS1_OAEP.new(private_key, hashAlgo=SHA256)
    decrypted_data = cipher.decrypt(data)
    return decrypted_data

def rsaSign(priKey: RSA.RsaKey, data: bytes) -> bytes:
    # private_key = RSA.import_key(priKey)
    h = SHA256.new(data)
    signature = pss.new(priKey).sign(h)
    return signature

def rsaVerify(pubKey, data, signature):
    public_key = RSA.import_key(pubKey)
    h = SHA256.new(data)
    verifier = pss.new(public_key)
    try:
        verifier.verify(h, signature)
        return True
    except (ValueError, TypeError):
        return False
    
def request(data, session: Session):
    headers = {}
    headers['Content-Type'] = 'application/json'
    headers['wa'] = session.walletAddress
    headers['chain'] = session.chainID
    headers['ver'] = settings.VERSION
    headers['app'] = session.appKey
    if session.token:
        headers['token'] = session.token
    response = session.requstSession.post(session.gateway, json=data, headers=headers)
    logging.debug("============== Mind SDK request ==============")
    logging.debug('MindSDKHeaders: %s'%headers)
    logging.debug("MindSDKData: %s"%data)
    logging.debug('MindSDKRequest: %s'%response.request.body.decode('utf-8'))
    logging.debug('MindSDKResponse: %s'%response.text)
    if response and response.status_code == 200:
        return json.loads(response.text)
    else:
        return None


import base64
import json
import logging
import struct
import uuid
from Crypto.Random import get_random_bytes
from Crypto.PublicKey import RSA
from web3 import Web3
from mindlakesdk import settings
import mindlakesdk.utils
import mindlakesdk.message
from base64 import b64decode, b64encode
from nacl.public import Box, PrivateKey, PublicKey

def registerMK(session: mindlakesdk.utils.Session):
    pubKey = session.nodePK.replace("\\n", "\n")
    ephemeralKey = get_random_bytes(16)
    sealedEphemeralKey = mindlakesdk.utils.rsaEncrypt(pubKey, ephemeralKey)
    sealedEphemeralKeyLenBytes = len(sealedEphemeralKey).to_bytes(2, 'little')
    iv = get_random_bytes(16)
    envelope_json = __genMKEnvelope(session.mk)
    envelope_enc = mindlakesdk.utils.aesEncrypt(ephemeralKey, iv, envelope_json.encode())
    big_envelope = sealedEphemeralKeyLenBytes + sealedEphemeralKey + iv + envelope_enc
    envelope = base64.b64encode(big_envelope).decode()
    return mindlakesdk.message.sendMKRegister(session, envelope)

def __genMKEnvelope(mk, accountID = None):
    envelope = {}
    if accountID:
        envelope["mekid"] = accountID
    base64MK = base64.b64encode(mk)
    envelope["mek"] = base64MK.decode()
    envelope["expire"] = 0
    envelope_json = json.dumps(envelope).replace(" ", "")
    return envelope_json

def registerPK(session: mindlakesdk.utils.Session):
    privateKey = session.sk
    publicKey = privateKey.publickey().export_key("PEM").decode()
    pkIDBytes = __genPKid(publicKey)
    pkIDStr = pkIDBytes.decode()
    session.pkID = pkIDStr
    toBeSignedBytes = bytearray(struct.pack("<q", session.accountID)) + pkIDBytes
    rsaSig = b'\x01' + mindlakesdk.utils.rsaSign(privateKey, toBeSignedBytes)
    rsaSigStr = base64.b64encode(rsaSig).decode()
    mkSig = b'\x00' + mindlakesdk.utils.hmacHash(session.mk, toBeSignedBytes)
    mkSigStr = base64.b64encode(mkSig).decode()
    return mindlakesdk.message.sendPKRegister(session, pkIDStr, publicKey, rsaSigStr, mkSigStr)

def __genPKid(pubKey):
    pubKeyHash = mindlakesdk.utils.sha256Hash(pubKey.encode("utf-8"))
    pubKeyHashBase64 = base64.b64encode(pubKeyHash)
    return pubKeyHashBase64

def encryptDKb64(mk, dkID, dk):
    dkID_dk = struct.pack('<Q', dkID) + dk
    iv = get_random_bytes(16)
    encrypted_data = mindlakesdk.utils.aesEncrypt(mk, iv, dkID_dk)
    dkCipher = b'\x03' + iv + encrypted_data
    return base64.b64encode(dkCipher).decode(encoding="ascii")

def decryptDKb64(mk, dkCipherStr):
    dkCipher = base64.b64decode(dkCipherStr)
    dkID_dk = mindlakesdk.utils.aesDecrypt(mk, dkCipher[1:17], dkCipher[17:])
    dkID = struct.unpack('<Q', dkID_dk[:8])[0]
    return dkID, dkID_dk[8:]

def genDK(session: mindlakesdk.utils.Session, table, column):
    result = mindlakesdk.message.getNextDKid(session)
    if not result:
        return result
    dkID = int(result.data)
    dk = get_random_bytes(16)
    dkCipherStr = encryptDKb64(session.mk, dkID, dk)
    grpID = uuid.uuid4()
    grpIDbytes = grpID.bytes
    grpIDStr = str(grpID)
    gAuthStr = __digest_gAuth(session.mk, grpIDbytes, dkID)
    return mindlakesdk.message.sendDKRegister(session, table, column, dkID, dkCipherStr, grpIDStr, gAuthStr)

def __digest_gAuth(mk, grpID, dkID):
    buf = grpID + struct.pack("<q", dkID)
    gAuth = mindlakesdk.utils.hmacHash(mk, buf)
    return base64.b64encode(gAuth).decode(encoding="ascii")

def prepareKeys(web3, walletAccount, chain: mindlakesdk.utils.BlockChain):
    mkCipher, skCipher = __loadKeysFromChain(walletAccount.address, web3, chain)
    logging.debug('mkCipher: %s'%mkCipher)
    logging.debug('skCipher: %s'%skCipher)
    
    if mkCipher and skCipher:
        logging.debug('Found keys in contract')
        mk = __decryptWithWalletKey(walletAccount, mkCipher)
        skIV, skCipher = skCipher[:16], skCipher[16:]
        skBytes = mindlakesdk.utils.aesDecrypt(mk, skIV, skCipher)
        sk = RSA.import_key(skBytes)
    else:
        logging.debug('No keys found in contract. Generate new keys')
        mk = mindlakesdk.utils.genAESKey()
        sk = mindlakesdk.utils.genRSAKey()
        skBytes = sk.exportKey('DER', pkcs=8)
        skIV = mindlakesdk.utils.get_random_bytes(16)
        skCipher = mindlakesdk.utils.aesEncrypt(mk, skIV, skBytes)
        skCipher = skIV + skCipher
        mkCipher = __encryptWithWalletKey(walletAccount, mk)
        __saveKeysToChain(web3, walletAccount, mkCipher, skCipher, chain)
    logging.debug('mk: %s'%mk)
    logging.debug('sk: %s'%sk)
    return mk, sk

def __saveKeysToChain(web3, walletAccount, MKCipher, SKCipher, chain: mindlakesdk.utils.BlockChain):
    walletAddress = walletAccount.address
    contract = web3.eth.contract(address=chain.contract, abi=chain.abi)
    nonce = web3.eth.get_transaction_count(walletAddress)
    gasEstimate = contract.functions.setKeys(MKCipher, SKCipher).estimate_gas({'from': walletAddress})
    txn = contract.functions.setKeys(MKCipher, SKCipher).build_transaction({
        'from': walletAddress,
        'gas': gasEstimate,
        'gasPrice': web3.eth.gas_price,
        'nonce': nonce,
    })
    signed_txn = walletAccount.signTransaction(txn)
    txn_hash = web3.eth.send_raw_transaction(signed_txn.rawTransaction)
    logging.debug(f'Transaction sent: {txn_hash.hex()}')

    txn_receipt = web3.eth.wait_for_transaction_receipt(txn_hash)
    logging.debug(f'Transaction mined, status: {txn_receipt["status"]}')

def __loadKeysFromChain(walletAddress: str, web3: Web3, chain: mindlakesdk.utils.BlockChain):
    contract = web3.eth.contract(address=chain.contract, abi=chain.abi)
    return contract.functions.getKeys(walletAddress).call()

def __encryptWithWalletKey(walletAccount, msg):
    ephemeralPrivKey = PrivateKey.generate()
    privKey = PrivateKey(walletAccount.key)
    pubKey = privKey.public_key
    encryptBox = Box(ephemeralPrivKey, pubKey)
    nounce = get_random_bytes(24)
    msg = b64encode(msg)
    encryptedMsg = encryptBox.encrypt(msg, nounce)
    return bytes(ephemeralPrivKey.public_key) + nounce + encryptedMsg.ciphertext

def __decryptWithWalletKey(walletAccount, data):
    privKey = PrivateKey(walletAccount.key)
    ephemeralPubKey, nounce, ciphertext = data[:32], data[32:56], data[56:]
    box = Box(privKey, PublicKey(ephemeralPubKey))
    return b64decode(box.decrypt(ciphertext, nounce))


import base64
import datetime
import json
import logging
import uuid

from mindlakesdk.utils import ResultType, Session
import mindlakesdk.utils
import mindlakesdk.message

class Permission:
    def __init__(self, session: Session):
        self.__session = session

    def grant(self, targetWalletAddress: str, columns: list, targetChainID: str = None) -> ResultType:
        session = self.__session
        if not targetChainID:
            targetChainID = session.chainID
        result = mindlakesdk.message.getPKidByWalletAddress(session, targetWalletAddress, targetChainID)
        if not result:
            return result
        targetPKID = result.data["publicKeyId"]
        logging.debug("Get PK ID from wallet address: " + targetWalletAddress + ' - ' + targetPKID)
        policy = {
            'issuer_dek_group': [],
            'subject_dek_group': []
        }
        for column in columns:
            tableName, columnName = column.split('.')
            result = mindlakesdk.message.getDKbyName(session, session.walletAddress, tableName, columnName)
            if not result:
                return result
            groupID = result.data['groupId']
            policy['issuer_dek_group'].append({
                'groupid': groupID,
                'min': 1,
                'max': 1000
            })
        result = self.__createPolicyBody(self.__session.pkID, targetPKID, policy)
        if not result:
            return result
        policyBodyJson = result.data
        signature = self.__signPolicy(policyBodyJson)
        signatureB64 = base64.b64encode(signature).decode('utf-8')
        return mindlakesdk.message.sendGrant(self.__session, policyBodyJson, signatureB64)

    def confirm(self, policyID: str) -> ResultType:
        result = mindlakesdk.message.getPolicyBySerialNumber(self.__session, policyID)
        if not result:
            return result
        policyBodyJson = result.data
        policyBody = json.loads(policyBodyJson)
        if policyBody['subject_pukid'] != self.__session.pkID:
            logging.debug("Self pukid: " + self.__session.pkID)
            logging.debug("Policy pukid: " + policyBody['subject_pukid'])
            return mindlakesdk.utils.ResultType(60002, "The policy is not for you")
        signature = self.__signPolicy(policyBodyJson)
        signatureB64 = base64.b64encode(signature).decode('utf-8')
        return mindlakesdk.message.sendConfirm(self.__session, policyBodyJson, signatureB64)
    
    def revoke(self, targetWalletAddress: str, targetChainID: str = None) -> ResultType:
        return self.grant(targetWalletAddress, [], targetChainID)

    def grantToSelf(self):
        result = mindlakesdk.message.getDKbyName(self.__session)
        if not result:
            return result
        groupID = result.data['groupId']
        policy = {
            'issuer_dek_group': [{
                'groupid': groupID,
                'min': 1,
                'max': 1000
            }],
            'subject_dek_group': [{
                'groupid': groupID,
                'min': 1,
                'max': 1000
            }]
        }
        result = self.__createPolicyBody(self.__session.pkID,
                                self.__session.pkID,
                                policy)
        if not result:
            return result
        policyBodyJson = result.data
        signature = self.__signPolicy(policyBodyJson)
        signatureB64 = base64.b64encode(signature).decode('utf-8')
        return mindlakesdk.message.sendSelfGrant(self.__session, policyBodyJson, signatureB64)

    def __createPolicyBody(self,
                            issuerPukid, 
                            subjectPukid,
                            policy,
                            version = 1,
                            serialNum = None,
                            notBefore = None,
                            notAfter = None,
                            resultDK = "SUBJECT",
                            operation = ['*'],
                            postProc = "NULL",
                            preProc = "NULL") -> ResultType:
        result = mindlakesdk.message.getPolicyByPKid(self.__session, issuerPukid, subjectPukid)
        if not result:
            return result
        policyBodyJson = result.data
        if policyBodyJson:
            logging.debug("Policy already exists, loading existing Policy")
            policyBody = json.loads(policyBodyJson)
        else:
            if not serialNum:
                serialNum = str(uuid.uuid4())
            if not notBefore:
                notBefore = datetime.datetime.now()
            if not notAfter:
                notAfter = datetime.datetime.now() + datetime.timedelta(days=365)
            policyBody = {}
            policyBody['version'] = version
            policyBody['serial_num'] = serialNum
            policyBody['issuer_pukid'] = issuerPukid
            policyBody['subject_pukid'] = subjectPukid
            policyBody['validity'] = {}
            policyBody['validity']['not_after'] = notAfter.astimezone().strftime('%Y%m%d%H%M%S%z')
            policyBody['validity']['not_before'] = notBefore.astimezone().strftime('%Y%m%d%H%M%S%z')
        policyBody['policies'] = {}
        policyBody['policies']['operation'] = operation
        policyBody['policies']['post_proc'] = postProc
        policyBody['policies']['pre_proc'] = preProc
        policyBody['policies']['result_dek'] = resultDK
        policyBody['policies'].update(policy)
        policyBodyJson = json.dumps(policyBody)
        return ResultType(0, "Success", policyBodyJson)
    
    def __signPolicy(self, policyBodyJson) -> bytes:
        toBeSignedBytes = policyBodyJson.encode('utf-8')
        signature = mindlakesdk.utils.rsaSign(self.__session.sk, toBeSignedBytes)
        return b'\x01' + signature
    
    def listGrantee(self) -> ResultType:
        return mindlakesdk.message.sendListGrantee(self.__session)  
    
    def listGrantedColumn(self, walletAddress: str, targetChainID: str = None) -> ResultType:
        if not targetChainID:
            targetChainID = self.__session.chainID
        return mindlakesdk.message.sendListGrantedColumn(self.__session, walletAddress, targetChainID) 
    
    def listOwner(self) -> ResultType:
        return mindlakesdk.message.sendListOwner(self.__session)
    
    def listOwnerColumn(self, walletAddress: str, targetChainID: str = None) -> ResultType:
        if not targetChainID:
            targetChainID = self.__session.chainID
        return mindlakesdk.message.sendListOwnerColumn(self.__session, walletAddress, targetChainID)
    


import datetime
from decimal import Decimal
import struct
from enum import Enum
from Crypto.Random import get_random_bytes
from mindlakesdk.utils import ResultType, Session, DataType
import mindlakesdk.message
import mindlakesdk.keyhelper
import mindlakesdk.utils

class Cryptor:

    class EncType(Enum):
        enc_int4 = 1
        enc_int8 = 2
        enc_float4 = 3
        enc_float8 = 4
        enc_decimal = 6
        enc_text = 7
        enc_timestamp = 8

    def __init__(self, session: Session):
        self.__session = session
        self.cryptParamsByColumn = {}
        self.cryptParamsByID = {}
    
    def encrypt(self, data, columnOrType) -> ResultType:
        result = self.__getEncryptParams(columnOrType)
        if not result:
            return result
        ctxid, encTypeNum, dk, alg, dataType = result.data
        data = Cryptor.__encodeByDataType(data, dataType)
        header = Cryptor.__genCryptoHeader(ctxid, encTypeNum)
        checkCode = Cryptor.__genCheckCode(data, 1)
        data_to_enc = data + checkCode
        if alg == 3:
            iv = get_random_bytes(16)
            encrypted_data = mindlakesdk.utils.aesEncrypt(dk, iv, data_to_enc)
        elif alg == 0:
            iv = get_random_bytes(12)
            encrypted_data = mindlakesdk.utils.aesGCMEncrypt(dk, iv, data_to_enc)
        buf = header + iv + encrypted_data
        tmp = buf[1:]
        checkCode2 = Cryptor.__genCheckCode(tmp, 1)
        result = checkCode2 + tmp
        resultHex = '\\x' + result.hex()
        return ResultType(0, None, resultHex)

    def __getEncryptParams(self, columnOrType):
        result = self.cryptParamsByColumn.get(columnOrType)
        if not result:
            if isinstance(columnOrType, DataType):
                dataType = columnOrType
                result = mindlakesdk.message.getDKbyName(self.__session)
                if not result:
                    return result
            else:
                tableName, columnName = columnOrType.split('.')
                result = mindlakesdk.message.getDataTypeByName(self.__session, tableName, columnName)
                if not result:
                    return result
                dataType = DataType(result.data)
                result = mindlakesdk.message.getDKbyName(self.__session, self.__session.walletAddress, tableName, columnName)
                # Temporary solution for MS not returning Error Code
                if result.code == 40010:
                    # DK not found, create one
                    result = mindlakesdk.keyhelper.genDK(self.__session, tableName, columnName)
                    if not result:
                        return result
                elif not result:
                    return result
                else:
                    pass
            encTypeNum = Cryptor.EncType['enc_' + dataType.name].value
            ctxid = result.data['ctxId']
            dkCipher = result.data['encryptedDek']
            dkID, dk = mindlakesdk.keyhelper.decryptDKb64(self.__session.mk, dkCipher)
            alg = result.data['algorithm']
            result = (ctxid, encTypeNum, dk, alg, dataType)
            self.cryptParamsByColumn[columnOrType] = result
            self.cryptParamsByID[ctxid] = (dk, alg)
        return ResultType(0, None, result)


    def __encodeByDataType(data, dataType: DataType) -> bytes:
        if dataType == DataType.int4:
            result = struct.pack("<i", data)
        elif dataType == DataType.int8:
            result = struct.pack("<q", data)
        elif dataType == DataType.float4:
            result = struct.pack("<f", data)
        elif dataType == DataType.float8:
            result = struct.pack("<d", data)
        elif dataType == DataType.decimal:
            val = Decimal(data)
            val_str = str(val)
            result = val_str.encode('utf-8')
        elif dataType == DataType.text:
            result = data.encode('utf-8')
        elif dataType == DataType.timestamp:
            u_sec = int(data.timestamp() * 1000000)
            u_sec -= 946684800000000
            u_sec += int(datetime.datetime.now().astimezone().utcoffset().total_seconds() * 1000000)
            result = struct.pack('<q', u_sec)
        else:
            raise Exception("Unsupported encryption type")
        return result

    def decrypt(self, cipher) -> ResultType:
        if isinstance(cipher, str):
            data = bytes.fromhex(cipher[2:])
        else:
            data = cipher
        header = Cryptor.__extractCryptoHeader(data)
        encTypeNum = Cryptor.__extractEncType(header)
        ctxId = Cryptor.__extractCtxId(header)
        result = self.__getDecryptParams(ctxId)
        if not result:
            return result
        dk, alg = result.data
        idx = (header[1] & 0x7) + 2
        if alg == 3:
            iv = data[idx:idx+16]
            cipherBlob = data[idx+16:]
            plainBlob = mindlakesdk.utils.aesDecrypt(dk, iv, cipherBlob)
        elif alg == 0:
            iv = data[idx:idx+12]
            idx += 12
            mac = data[idx:idx+16]
            idx += 16
            cipherBlob = data[idx:]
            plainBlob = mindlakesdk.utils.aesGCMDecrypt(dk, iv, cipherBlob, mac)
        else:
            return ResultType(60004, "Unsupported algorithm to decrypt")
        result = plainBlob[:-1]
        checkCode = plainBlob[-1:]
        checkCode2 = Cryptor.__genCheckCode(result, 1)
        if checkCode != checkCode2:
            return ResultType(60005, "Check code of cipher is not correct")
        result = Cryptor.__decodeByEncType(result, Cryptor.EncType(encTypeNum))
        return ResultType(0, None, result)

    def __getDecryptParams(self, ctxId: int):
        params = self.cryptParamsByID.get(ctxId)
        if not params:
            result = mindlakesdk.message.getDKbyCid(self.__session, ctxId)
            if not result:
                return result
            if not result.data:
                return ResultType(60002, "Can't get data key")
            dkCipher = result.data['encryptedDek']
            if not dkCipher:
                return ResultType(60002, "Can't get data key")
            try:
                dkID, dk = mindlakesdk.keyhelper.decryptDKb64(self.__session.mk, dkCipher)
            except:
                return ResultType(60003, "Can't handle data key")
            alg = result.data['algorithm']
            if alg is None:
                return ResultType(60002, "Can't get data key")
            params = (dk, alg)
            self.cryptParamsByID[ctxId] = params
        return ResultType(0, None, params)

    def __decodeByEncType(data, encType: EncType):
        if encType == Cryptor.EncType.enc_int4:
            size = struct.calcsize('<i')
            buf = data[:size]
            result = struct.unpack('<i', buf)[0]
        elif encType == Cryptor.EncType.enc_int8:
            size = struct.calcsize('<q')
            buf = data[:size]
            result = struct.unpack('<q', buf)[0]
        elif encType == Cryptor.EncType.enc_float4:
            size = struct.calcsize('<f')
            buf = data[:size]
            result = struct.unpack('<f', buf)[0]
        elif encType == Cryptor.EncType.enc_float8:
            size = struct.calcsize('<d')
            buf = data[:size]
            result = struct.unpack('<d', buf)[0]
        elif encType == Cryptor.EncType.enc_decimal:
            result = Decimal(data.decode('utf-8'))
        elif encType == Cryptor.EncType.enc_text:
            result = data.decode('utf-8')
        elif encType == Cryptor.EncType.enc_timestamp:
            size = struct.calcsize('<q')
            buf = data[:size]
            u_sec = struct.unpack('<q', buf)[0]
            u_sec += 946684800000000
            u_sec -= int(datetime.datetime.now().astimezone().utcoffset().total_seconds() * 1000000)
            time_stamp = u_sec / 1000000.0
            result = datetime.datetime.fromtimestamp(time_stamp)
        else:
            raise Exception("Unsupported encryption type")
        return result

    def __extractCryptoHeader(data):
        header = bytearray()
        index = 0
        for i in range(1):
            header.append(data[index])
            index += 1
        assert index == 1
        header.append(data[index])
        index += 1
        rng = header[1] & 0x7
        for i in range(rng):
            header.append(data[index])
            index += 1
        return header
    
    def __extractEncType(header):
        tmp_value = header[1]
        type_value = (tmp_value & 0xF8) >> 3
        return type_value
    
    def __extractCtxId(header) -> int:
        ctxIdLen = header[1] & 0x7
        assert len(header) == ctxIdLen + 2
        ctxId = 0
        for i in range(ctxIdLen):
            index = len(header) - 1 - i
            ctxId = ctxId << 8 | (header[index] & 0xFF)
        return ctxId

    def __genCryptoHeader(ctxid, encType):
        head = bytearray(bytes.fromhex('0000'))

        tmp_value = head[1]
        tmp_value = tmp_value & 0xFFFFFF07
        tmp_value = tmp_value | (encType << 3)
        head[1] = tmp_value

        tmp = ctxid
        while tmp != 0:
            head.append(tmp & 0xFF)
            tmp >>= 8

        ctxLen = len(head) - 2

        tmp_val = head[1]
        tmp_val = (tmp_val & 0xFFFFFFF8) | (ctxLen & 0x7)
        head[1] = tmp_val
        return bytes(head)

    def __genCheckCode(data: bytes, resultSize):
        tmpCode = bytearray(resultSize)
        for i in range(len(data)):
            n = i % resultSize
            tmpCode[n] ^= data[i]
        return bytes(tmpCode)
    


from mindlakesdk.utils import ResultType, Session, DataType
import mindlakesdk.message
import logging

class DataLake:
    def __init__(self, session: Session):
        self.__session = session

    class Column(dict):
        def __init__(self, columnName: str, dataType: DataType, encrypt: bool):
            self.columnName = columnName
            self.type = dataType
            self.encrypt = encrypt
        
        def toDict(self):
            return {
                'columnName': self.columnName,
                'type': self.type.value,
                'encrypt': self.encrypt
            }

    def createTable(self, tableName: str, columns: list, primaryKey: list = None) -> ResultType:
        columnsDict = []
        for column in columns:
            columnsDict.append(column.toDict())
        return mindlakesdk.message.sendCreateTable(self.__session, tableName, columnsDict, primaryKey)
    
    def listCocoon(self) -> ResultType:
        logging.debug("listCocoon")
        return mindlakesdk.message.sendListCocoon(self.__session)
    
    def linkTableToCocoon(self, tableName: str, cocoonName: str) -> ResultType:
        return mindlakesdk.message.sendLinkTableToCocoon(self.__session, tableName, cocoonName) 
    
    def listTablesByCocoon(self, cocoonName: str) -> ResultType:
        return mindlakesdk.message.sendListTablesByCocoon(self.__session, cocoonName) 
    
    def createCocoon(self, cocoonName: str) -> ResultType:
        return mindlakesdk.message.sendCreateCocoon(self.__session, cocoonName) 
    
    def query(self, executeSql: str) -> ResultType:
        return mindlakesdk.message.sendQuery(self.__session, executeSql)
    
    def dropCocoon(self, cocoonName: str) -> ResultType:
        return mindlakesdk.message.sendDropCocoon(self.__session, cocoonName)
    
    def dropTable(self, tableName: str) -> ResultType:
        return mindlakesdk.message.sendDropTable(self.__session, tableName)  
    
    def queryForDataAndMeta(self, executeSql: str) -> ResultType:
        return mindlakesdk.message.sendQueryForDataAndMeta(self.__session, executeSql)



name = "mindlakesdk"

from eth_account.messages import encode_defunct
from web3 import Web3
import requests

import mindlakesdk.settings as settings
import mindlakesdk.utils
from mindlakesdk.utils import ResultType, Session, DataType, BlockChain
import mindlakesdk.keyhelper
from mindlakesdk.datalake import DataLake
from mindlakesdk.cryptor import Cryptor
from mindlakesdk.permission import Permission
import mindlakesdk.message

import logging

class MindLake(ResultType):

    DataType = DataType
    
    def __init__(self, walletPrivateKey: str, appKey: str, chainID: str = settings.DEFAULT_CHAINID, gateway: str = None):
        logging.debug(__name__)
        self.__session = mindlakesdk.utils.Session()
        session = self.__session
        session.requstSession = requests.Session()
        self.datalake = DataLake(session)
        self.cryptor = Cryptor(session)
        self.permission = Permission(session)

        session.chainID = chainID
        session.appKey = appKey
        if gateway:
            session.gateway = gateway
        else:
            session.gateway = settings.GATEWAY
        logging.debug('gateway: %s'%session.gateway)

        result = mindlakesdk.message.getChainInfo(session)
        if not result:
            self.__setResult(result)
            return
        chainInfo = result.data
        logging.debug('getChainInfo: %s'%result.data)
        chain = None
        for chainObj in chainInfo:
            if chainObj["chainId"] == chainID:
                chain = BlockChain(chainObj)
                break
        if not chain:
            self.__setResult(ResultType(60016, "Specified chainID not supported", None))
            return

        web3 = Web3(Web3.HTTPProvider(chain.rpcNode))
        walletAccount = web3.eth.account.from_key(walletPrivateKey)
        
        session.walletAddress = walletAccount.address
        logging.debug('walletAddress: %s'%session.walletAddress)

        result = mindlakesdk.message.getNounce(session)
        if not result:
            self.__setResult(result)
            return 
        nounce = result.data
        logging.debug('getNounce: %s'%nounce)

        result = MindLake.__login(session, walletAccount, nounce)
        if not result:
            self.__setResult(result)
            return 
        logging.debug('__login: %s'%result.data)
            
        session.mk, session.sk = mindlakesdk.keyhelper.prepareKeys(web3, walletAccount, chain)
        logging.debug('getNounce: %s'%nounce)

        result = mindlakesdk.message.getAccountInfo(session)
        if not result:
            self.__setResult(result)
            return 

        result = mindlakesdk.keyhelper.registerMK(session)
        if not result:
            self.__setResult(result)
            return 

        if not session.isRegistered:
            result = MindLake.__registerAccount(session, self.permission)
            if not result:
                self.__setResult(result)
                return 
        else:
            result = mindlakesdk.message.getPKid(session)
            if not result:
                self.__setResult(result)
                return 
        self.code = 0
        self.message = "Success"
        self.data = None

    @staticmethod
    def __login(session: Session, walletAccount, nounce):
        msgToSign = encode_defunct(text=nounce)
        signature = walletAccount.sign_message(msgToSign)
        signatureHex = signature.signature.hex()
        return mindlakesdk.message.sendLogin(session, signatureHex)

    @staticmethod
    def __registerAccount(session: Session, permission: Permission):
        result = mindlakesdk.keyhelper.registerPK(session)
        if not result:
            return result
        result = permission.grantToSelf()
        return result
    
    def __setResult(self, result):
        self.code = result.code
        self.message = result.message
        self.data = result.data

    def getNameSpace(self, walletAddress: str, chainID: str = None):
        if chainID is None:
            chainID = self.__session.chainID
        if chainID == settings.CLERK_CHAINID:
            return walletAddress.replace('@', '_').replace('.', '_')
        else:
            addressHex = walletAddress[2:].lower()
            if chainID == settings.DEFAULT_CHAINID:
                return addressHex
            else:
                return chainID + "_" + addressHex

connect = mindlakesdk.MindLake
    


from mindlakesdk.utils import ResultType, request, Session
import logging

def getNounce(session: Session):
    data = {"bizType":203}
    result = request(data, session)
    if result:
        return ResultType(result["code"], result["message"], result["data"])
    else:
        return ResultType(60001, "Network error", None)


def sendLogin(session: Session, signatureHex):
    data = {"bizType":201,"walletAddress":session.walletAddress,"signature":signatureHex}
    result = request(data, session)
    if result:
        if result["code"] == 0:
            # TODO: add exception handling
            session.isLogin = True
            session.token = result["data"]["token"]
        return ResultType(result["code"], result["message"], result["data"])
    else:
        return ResultType(60001, "Network error", None)
    
def getChainInfo(session: Session):
    data = {"bizType":205}
    result = request(data, session)
    if result:
        return ResultType(result["code"], result["message"], result["data"])
    else:
        return ResultType(60001, "Network error", None)
    
def getAccountInfo(session: Session):
    data = {"bizType":120}
    result = __requestCommon(session, data)
    if result:
        session.isRegistered = result.data["isRegistered"] and result.data["isMekProvision"] and result.data["isSelfBcl"]
        session.accountID = int(result.data["mekId"])
        session.nodePK = result.data["publicKey"]
    return result
    
def getPKid(session: Session):
    data = {"bizType":103,"mekId":session.accountID}
    result = __requestCommon(session, data)
    if result:
        session.pkID = result.data
    return result
    
def sendMKRegister(session: Session, envelope):
    data = {"bizType":102, "envelope":envelope, "databasePublicKey":session.nodePK}
    return __requestCommon(session, data)
    
def sendPKRegister(session: Session, pkIDStr, publicKey, rsaSigStr, mkSigStr):
    data = {
        "bizType":104, 
        "mekId":session.accountID, 
        "pukId":pkIDStr,
        "publicKey":publicKey,
        "privateSig":rsaSigStr,
        "mekSig":mkSigStr
    }
    return __requestCommon(session, data)

def getDKbyName(session: Session, walletAddress: str = None, 
                table: str = None, column: str = None) -> ResultType:
    data = {"bizType":108, "walletAddress":walletAddress, "table":table, "column":column}
    return __requestCommon(session, data)

def getNextDKid(session: Session) -> ResultType:
    data = {"bizType":109, "mekId":session.accountID}
    return __requestCommon(session, data)

def getDKbyCid(session: Session, cid: int) -> ResultType:
    data = {"bizType":111, "ctxId":cid}
    return __requestCommon(session, data)

def sendDKRegister(session: Session, table, column, dkID, dkCipherStr, grpIDStr, gAuthStr) -> ResultType:
    data = {
        "bizType":110, 
        "mekId":session.accountID, 
        "table":table,
        "column":column,
        "dekId":dkID,
        "dekCipherStr":dkCipherStr,
        "grpIdStr":grpIDStr,
        "groupAuthStr":gAuthStr
    }
    return __requestCommon(session, data)

def getDataTypeByName(session: Session, tableName: str, columnName: str) -> ResultType:
    data = {"bizType":107, "tableName":tableName, "column":columnName}
    return __requestCommon(session, data)

def getPolicyByPKid(session: Session, issuePKid: str, subjectPKid: str) -> ResultType:
    data = {"bizType":118, "issuePukId":issuePKid, "subjectPukId":subjectPKid}
    return __requestCommon(session, data)

def getPolicyBySerialNumber(session: Session, serialNumber: str) -> ResultType:
    data = {"bizType":116, "serialNum":serialNumber}
    return __requestCommon(session, data)

def getPKidByWalletAddress(session: Session, walletAddress: str, targetChainID: str) -> ResultType:
    data = {"bizType":119, "targetWalletAddress":walletAddress, "targetChain":targetChainID}
    return __requestCommon(session, data)

def sendGrant(session: Session, policyBodyJson: str, signature: str) -> ResultType:
    data = {"bizType":115, "bclBody":policyBodyJson, "privateSig":signature}
    return __requestCommon(session, data)

def sendConfirm(session: Session, policyBodyJson: str, signature: str) -> ResultType:
    data = {"bizType":117, "bclBody":policyBodyJson, "privateSig":signature}
    return __requestCommon(session, data)

def sendSelfGrant(session: Session, policyBodyJson: str, signature: str) -> ResultType:
    data = {"bizType":106, "bclBody":policyBodyJson, "privateSig":signature}
    return __requestCommon(session, data)

def sendCreateTable(session: Session, tableName: str, columns: list, primaryKey: list = None) -> ResultType:
    data = {"bizType":123, "tableName":tableName, "columns":columns, "pkColumns":primaryKey}
    return __requestCommon(session, data)

def sendListCocoon(session: Session) -> ResultType:
    data = {"bizType":122}
    return __requestCommon(session, data)

def sendListTablesByCocoon(session: Session, cocoonName: str) -> ResultType:
    data = {"bizType":125, 'cocoonName': cocoonName}
    return __requestCommon(session, data)

def sendLinkTableToCocoon(session: Session, tableName: str, cocoonName: str) -> ResultType:
    data = {"bizType":124, 'tableName': tableName, 'cocoonName': cocoonName}
    return __requestCommon(session, data) 

def sendCreateCocoon(session: Session, cocoonName: str) -> ResultType:
    data = {"bizType":121, 'cocoonName': cocoonName}
    return __requestCommon(session, data)

def sendQuery(session: Session, executeSql) -> ResultType:
    data = {"bizType":114, 'executeSql': executeSql}
    return __requestCommon(session, data)

def sendQueryForDataAndMeta(session: Session, executeSql) -> ResultType:
    data = {"bizType":113, 'executeSql': executeSql}
    return __requestCommon(session, data)

def sendDropCocoon(session: Session, cocoonName: str) -> ResultType:
    data = {"bizType":129, 'cocoonName': cocoonName}
    return __requestCommon(session, data)

def sendDropTable(session: Session, tableName: str) -> ResultType:
    data = {"bizType":128, 'tableName': tableName}
    return __requestCommon(session, data)

def sendListGrantee(session: Session) -> ResultType:
    data = {"bizType":126}
    return __requestCommon(session, data)

def sendListGrantedColumn(session: Session, targetWalletAddress: str, targetChainID: str) -> ResultType:
    data = {"bizType":127, 'targetWalletAddress': targetWalletAddress, 'targetChain': targetChainID}
    return __requestCommon(session, data)

def sendListOwner(session: Session) -> ResultType:
    data = {"bizType":130}
    return __requestCommon(session, data)

def sendListOwnerColumn(session: Session, targetWalletAddress: str, targetChainID: str) -> ResultType:
    data = {"bizType":131, 'targetWalletAddress': targetWalletAddress, 'targetChain': targetChainID}
    return __requestCommon(session, data)

def __requestCommon(session: Session, data: str) -> ResultType:
    if session.isLogin:
        result = request(data, session)
        if result:
            return ResultType(result["code"], result["message"], result["data"])
        else:
            return ResultType(60001, "Network error", None)
    else:
        return ResultType(40001, "Not logged in", None)